(in-package "MORPHICL")

(defclass canvas ()
  ((foreground-color   :type color :initarg :foreground-color :accessor canvas-foreground-color)
   (background-color   :type color     :initarg :background-color   :accessor canvas-background-color)
   (font               :type font      :initarg :font               :accessor canvas-font)
   (transformation     :type rectangle :initarg :transformation     :accessor canvas-transformation)
   (clipping-rectangle :type rectangle :initarg :clipping-rectangle :accessor canvas-clipping-rectangle)))

(defun canvasp (object)
  (typep object 'canvas))

(defgeneric canvas-width ((canvas canvas)))
(defgeneric canvas-height ((canvas canvas)))
(defgeneric canvas-transform-point ((canvas canvas) (point point)))
(defgeneric canvas-transform-rectangle ((canvas canvas) (rectangle rectangle)))
(defgeneric canvas-translate-point ((canvas canvas) (point point)))
(defgeneric canvas-translate-rectangle ((canvas canvas) (rectangle rectangle)))
(defgeneric canvas-draw-line ((canvas canvas) (start point) (end point)))
(defgeneric canvas-draw-lines ((canvas canvas) points))
(defgeneric canvas-draw-point ((canvas canvas) (point point)))
(defgeneric canvas-draw-points ((canvas canvas) points))
(defgeneric canvas-outline-polygon ((canvas canvas) points))
(defgeneric canvas-fill-polygon ((canvas canvas) points))
(defgeneric canvas-outline-rectangle ((canvas canvas) (rectangle rectangle)))
(defgeneric canvas-fill-rectangle ((canvas canvas) (rectangle rectangle))) 
(defgeneric canvas-outline-ellipse ((canvas canvas) (rectangle rectangle) &optional start end))
(defgeneric canvas-fill-ellipse ((canvas canvas) (rectangle rectangle) &optional start end))
(defgeneric canvas-draw-text ((canvas canvas) text (offset point) &key start end)) 

(defgeneric canvas-foreground-color ((canvas canvas)))
(defgeneric (setf canvas-foreground-color) ((color color) (canvas canvas)))
(defgeneric canvas-background-color ((canvas canvas)))
(defgeneric (setf canvas-background-color) ((color color) (canvas canvas)))
(defgeneric canvas-font ((canvas canvas)))
(defgeneric (setf canvas-font) ((font font) (canvas canvas)))
(defgeneric canvas-transformation ((canvas canvas)))
(defgeneric (setf canvas-transformation) ((rectangle rectangle) (canvas canvas)))
(defgeneric canvas-clipping-rectangle ((canvas canvas)))
(defgeneric (setf canvas-clipping-rectangle) ((rectangle rectangle) (canvas canvas)))

(defmacro with-canvas-options ((canvas 
                                &key foreground-color
                                     background-color
                                     font
                                     transformation
                                     clipping-rectangle)
                               &body body)
  (let ((canvas-var     (gensym))
        (foreground-var (gensym))
        (background-var (gensym))
        (font-var       (gensym))
        (transform-var  (gensym))
        (clipping-var   (gensym)))
    `(let* ((,canvas-var ,canvas)
            ,@(when foreground-color
                `((,foreground-var (canvas-foreground-color ,canvas-var))))
            ,@(when background-color
                `((,background-var (canvas-background-color ,canvas-var))))
            ,@(when font
                `((,font-var (canvas-font ,canvas-var))))
            ,@(when transformation
                `((,transform-var (canvas-transformation ,canvas-var))))
            ,@(when clipping-rectangle
                `((,clipping-var (canvas-clipping-rectangle ,canvas-var)))))
       (psetf
         ,@(when foreground-color
             `((canvas-foreground-color ,canvas-var) ,foreground-color))
         ,@(when background-color
             `((canvas-background-color ,canvas-var) ,background-color)) 
         ,@(when font
             `((canvas-font ,canvas-var) ,font))
         ,@(when transformation
             `((canvas-transformation ,canvas-var) ,transformation))
         ,@(when clipping-rectangle
             `((canvas-clipping-rectangle ,canvas-var) ,clipping-rectangle)))
       (prog1 
         (progn ,@body)
         (psetf
           ,@(when foreground-color
               `((canvas-foreground-color ,canvas-var) ,foreground-var))
           ,@(when background-color
               `((canvas-background-color ,canvas-var) ,background-var)) 
           ,@(when font
               `((canvas-font ,canvas-var) ,font-var))
           ,@(when transformation
               `((canvas-transformation ,canvas-var) ,transform-var))
           ,@(when clipping-rectangle
               `((canvas-clipping-rectangle ,canvas-var) ,clipping-var)))))))

(defmethod canvas-transform-point ((canvas canvas) (point point))
  (let ((transformation (canvas-transformation canvas)))
    (point-round
      (point-translate 
        (point-scale point
          (rectangle-width transformation)
          (rectangle-height transformation))
        (rectangle-left transformation)
        (rectangle-top transformation)))))

(defmethod canvas-transform-rectangle ((canvas canvas) (rectangle rectangle))
  (let ((transformation (canvas-transformation canvas)))
    (rectangle-round
      (rectangle-translate
        (rectangle-scale rectangle
          (rectangle-width transformation)
          (rectangle-height transformation))
        (rectangle-left transformation)
        (rectangle-top transformation)))))

(defmethod canvas-translate-point ((canvas canvas) (point point))
  (let ((transformation (canvas-transformation canvas)))
    (point-translate point 
      (rectangle-left transformation)
      (rectangle-top transformation))))

(defmethod canvas-translate-rectangle ((canvas canvas) (rectangle rectangle))
  (let ((transformation (canvas-transformation canvas)))
    (rectangle-translate rectangle
      (rectangle-left transformation)
      (rectangle-top transformation))))

